
// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Javier Sánchez Pérez <jsanchez@dis.ulpgc.es>
// All rights reserved.

#ifndef BICUBIC_INTERPOLATION_H
#define BICUBIC_INTERPOLATION_H


#include <iostream>
#include <assert.h>

#include "bilinear_interpolation.h"

/**
  *
  * Bicubic interpolation in one dimension
  *
**/
double cubic_interpolation (
    double v[4],  ///interpolation points
    double x      ///point to be interpolated
)
{
	return v[1] + 0.5 * x*(v[2] - v[0]
                + x*(2.0*v[0] - 5.0*v[1] + 4.0*v[2] - v[3]
                + x*(3.0*(v[1] - v[2]) + v[3] - v[0])));
}

/**
  *
  * Bicubic interpolation in two dimension
  *
**/
double bicubic_interpolation (
    double p[4][4], ///array containing the interpolation points
    double x,       ///x position to be interpolated
    double y        ///y position to be interpolated
)
{
	double v[4];
	v[0] = cubic_interpolation(p[0], y);
	v[1] = cubic_interpolation(p[1], y);
	v[2] = cubic_interpolation(p[2], y);
	v[3] = cubic_interpolation(p[3], y);
	return cubic_interpolation(v, x);
}

/**
  *
  * Calculate the bicubic interpolation of a point in
  * an image. It also detects if the point goes beyond the
  * image limits
  *
**/
float bicubic_interpolation(
    const float *input, ///image to be interpolated
    const float uu,     ///x component of the vector field
    const float vv,     ///y component of the vector field
    const int   nx,     ///width of the image
    const int   ny,     ///height of the image
    float       &mask   ///mask to detect the motions outside the image
)
{
    const int sx = (uu < 0)? -1: 1;
    const int sy = (vv < 0)? -1: 1;

    int x, y, mx, my, dx, dy, ddx, ddy;

    switch(BOUNDARY_CONDITION) {

    case 0:x   = neumann_bc((int) uu, nx, mask);
           y   = neumann_bc((int) vv, ny, mask);
           mx  = neumann_bc((int) uu - sx, nx, mask);
           my  = neumann_bc((int) vv - sx, ny, mask);
           dx  = neumann_bc((int) uu + sx, nx, mask);
           dy  = neumann_bc((int) vv + sy, ny, mask);
           ddx = neumann_bc((int) uu + 2*sx, nx, mask);
           ddy = neumann_bc((int) vv + 2*sy, ny, mask);
           break;

    case 1:x   = periodic_bc((int) uu, nx, mask);
           y   = periodic_bc((int) vv, ny, mask);
           mx  = periodic_bc((int) uu - sx, nx, mask);
           my  = periodic_bc((int) vv - sx, ny, mask);
           dx  = periodic_bc((int) uu + sx, nx, mask);
           dy  = periodic_bc((int) vv + sy, ny, mask);
           ddx = periodic_bc((int) uu + 2*sx, nx, mask);
           ddy = periodic_bc((int) vv + 2*sy, ny, mask);
           break;

    case 2:x   = symmetric_bc((int) uu, nx, mask);
           y   = symmetric_bc((int) vv, ny, mask);
           mx  = symmetric_bc((int) uu - sx, nx, mask);
           my  = symmetric_bc((int) vv - sx, ny, mask);
           dx  = symmetric_bc((int) uu + sx, nx, mask);
           dy  = symmetric_bc((int) vv + sy, ny, mask);
           ddx = symmetric_bc((int) uu + 2*sx, nx, mask);
           ddy = symmetric_bc((int) vv + 2*sy, ny, mask);
           break;

    default:x   = neumann_bc((int) uu, nx, mask);
            y   = neumann_bc((int) vv, ny, mask);
            mx  = neumann_bc((int) uu - sx, nx, mask);
            my  = neumann_bc((int) vv - sx, ny, mask);
            dx  = neumann_bc((int) uu + sx, nx, mask);
            dy  = neumann_bc((int) vv + sy, ny, mask);
            ddx = neumann_bc((int) uu + 2*sx, nx, mask);
            ddy = neumann_bc((int) vv + 2*sy, ny, mask);
            break;
    }

    //obtain the interpolation points of the image
    const float p11 = input[mx  + nx * my];
    const float p12 = input[x   + nx * my];
    const float p13 = input[dx  + nx * my];
    const float p14 = input[ddx + nx * my];

    const float p21 = input[mx  + nx * y];
    const float p22 = input[x   + nx * y];
    const float p23 = input[dx  + nx * y];
    const float p24 = input[ddx + nx * y];

    const float p31 = input[mx  + nx * dy];
    const float p32 = input[x   + nx * dy];
    const float p33 = input[dx  + nx * dy];
    const float p34 = input[ddx + nx * dy];

    const float p41 = input[mx  + nx * ddy];
    const float p42 = input[x   + nx * ddy];
    const float p43 = input[dx  + nx * ddy];
    const float p44 = input[ddx + nx * ddy];

    //create array
    double pol[4][4] = {{p11,p21,p31,p41}, {p12,p22,p32,p42}, {p13,p23,p33,p43}, {p14,p24,p34,p44}};

    return bicubic_interpolation(pol, (float)uu-x, (float)vv-y);
}


/**
  *
  * Calculate the bicubic interpolation of an image.
  * It also creates a mask for the points that go beyond the
  * image limits
  *
**/
void bicubic_interpolation(
    const float *input,  ///image to be warped
    const float *u,      ///x component of the vector field
    const float *v,      ///y component of the vector field
    float       *output, ///warped output image with bicubir interpol
    const int    nx,     ///width of the image
    const int    ny,     ///height of the image
    float       *mask    ///mask to detect the motions outside the image
)
{
    for(int i = 0; i < nx * ny; i++)
        mask[i] = 1;

 #pragma omp parallel
 {
    #pragma omp for schedule(dynamic) nowait
    for(int j = 0; j < ny; j++)

       for(int i = 0; i < nx; i++)
       {
          const int   p = i + nx * j;
          const float uu = (float) (i + u[p]);
          const float vv = (float) (j + v[p]);

          output[p] = bicubic_interpolation(input, uu, vv, nx, ny, mask[p]);
       }
 }
}


#endif
