
// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Javier Sánchez Pérez <jsanchez@dis.ulpgc.es>
// All rights reserved.


#include <algorithm>
#include <iostream>

#include "robust_optic_flow.h"
#include "file_uv.h"


extern "C" {
#include "iio.h"
}



using namespace std;



bool read_image(const char *fname, float **f, int *w, int *h)
{
	*f = iio_read_image_float(fname, w, h);
	return *f ? true : false;
}


int main(int argc, char *argv[])
{
    if(argc < 3)

        cout << "Usage: " << argv[0]
             << " I1 I2 [out_file processors dt"
             << " alpha gamma nscales zoom_factor"
             <<	" TOL inner_iter outer_iter verbose]"
             << endl;
	     
    else
    {
        int i = 1;

        //read the parameters
        const char *image1  = argv[i]; i++;
        const char *image2  = argv[i]; i++;
        const char *outfile = (argc >= 4)?  argv[i]:"flow.uv"; i++;

        int   nproc   = (argc >= 5)?  atoi(argv[i]):0;      i++;
        float dt      = (argc >= 6)?  atof(argv[i]):5;      i++;
        float alpha   = (argc >= 7)?  atof(argv[i]):10;     i++;
        float gamma   = (argc >= 8)?  atof(argv[i]):10;     i++;
        int   nscales = (argc >= 9)?  atoi(argv[i]):5;      i++;
        float zfactor = (argc >= 10)? atof(argv[i]):0.5;    i++;
        float TOL     = (argc >= 11)? atof(argv[i]):0.0001; i++;
        int   initer  = (argc >= 12)? atoi(argv[i]):10;     i++;
        int   outiter = (argc >= 13)? atoi(argv[i]):30;     i++;
        int   verbose = (argc >= 14)? atoi(argv[i]):0;      i++;

        //check parameters
        if(nproc   >  0) omp_set_num_threads(nproc);
        if(dt      <= 0) dt      = 10;
        if(alpha   <= 0) alpha   = 100;
        if(gamma   <= 0) gamma   = 100;
        if(nscales <= 0) nscales = 5;
        if(zfactor <  0.5 || zfactor >= 1) zfactor = 0.5;
        if(TOL     <= 0) TOL     = 0.0001;
        if(initer  <= 0) initer  = 10;
        if(outiter <= 0) outiter = 30;

        int    nx, ny;
        float *I1;  //the source image
	float *I2;  //the target image
        
        bool correct1 = read_image(image1, &I1, &nx, &ny);
        bool correct2 = read_image(image2, &I2, &nx, &ny);

        if(correct1 && correct2)
        {  
            const float N = 1 + log(std::min(nx, ny) / 8.) / log(1./zfactor);

            if((int) N < nscales) nscales = (int) N;

            cout  << endl << "ncores:" << nproc << " dt:" << dt << " alpha:" << alpha << " gamma:" << gamma
                  << " scales:" << nscales << " nu:" << zfactor << " TOL:" << TOL << " inner:" << initer
                  << " outer:" << outiter << endl;
		  
	    //allocate memory for the flow
	    float *u = new float[nx*ny];
	    float *v = new float[nx*ny];

	    //call the optic flow method
	    optic_flow_zoom_robust(
		I1, I2, u, v, nx, ny, dt,
		alpha, gamma, nscales, zfactor,
		TOL, initer, outiter, verbose
	    );

	    //save the optic flow in our .uv file format
	    save_uv(u, v, nx, ny, "flow.uv");

	    //free dynamic memory
	    delete []I1;
	    delete []I2;
	    delete []u;
	    delete []v;
	}
    }

    return 0;
}

