
// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Javier Sánchez Pérez <jsanchez@dis.ulpgc.es>
// All rights reserved.

#ifndef MASK_H
#define MASK_H

#include <omp.h>

/**
 *
 * Function to compute the gradient with centered differences
 *
 */
void gradient(
    const float *input, ///input image
    float *dx,          ///computed x derivative
    float *dy,          ///computed y derivative
    const int nx,       ///image width
    const int ny        ///image height
)
{
 #pragma omp parallel
 {
    //gradient in the center body of the image
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        for(int j = 1; j < nx-1; j++)
        {
            const int k = i * nx + j;
            dx[k] = 0.5*(input[k+1] - input[k-1]);
            dy[k] = 0.5*(input[k+nx] - input[k-nx]);
        }
    }

    //gradient in the first and last rows
    #pragma omp for schedule(dynamic) nowait
    for(int j = 1; j < nx-1; j++)
    {
        dx[j] = 0.5*(input[j+1] - input[j-1]);
        dy[j] = 0.5*(input[j+nx] - input[j]);

        const int k = (ny - 1) * nx + j;

        dx[k] = 0.5*(input[k+1] - input[k-1]);
        dy[k] = 0.5*(input[k] - input[k-nx]);
    }

    //gradient in the first and last columns
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        const int p = i * nx;
        dx[p] = 0.5*(input[p+1] - input[p]);
        dy[p] = 0.5*(input[p+nx] - input[p-nx]);

        const int k = (i+1) * nx - 1;

        dx[k] = 0.5*(input[k] - input[k-1]);
        dy[k] = 0.5*(input[k+nx] - input[k-nx]);
    }

    //calculate the gradient in the corners
    dx[0] = 0.5*(input[1] - input[0]);
    dy[0] = 0.5*(input[nx] - input[0]);

    dx[nx-1] = 0.5*(input[nx-1] - input[nx-2]);
    dy[nx-1] = 0.5*(input[2*nx-1] - input[nx-1]);

    dx[(ny-1)*nx] = 0.5*(input[(ny-1)*nx + 1] - input[(ny-1)*nx]);
    dy[(ny-1)*nx] = 0.5*(input[(ny-1)*nx] - input[(ny-2)*nx]);

    dx[ny*nx-1] = 0.5*(input[ny*nx-1] - input[ny*nx-1-1]);
    dy[ny*nx-1] = 0.5*(input[ny*nx-1] - input[(ny-1)*nx-1]);
 }

}


/**
 *
 * Function to apply a 3x3 mask to an image
 *
 */
void mask3x3(
    const float *input, ///input image
    float *output,      ///output image
    const int nx,       ///image width
    const int ny,       ///image height
    const float *mask   ///mask to be applied
)
{

 #pragma omp parallel
 {
    //apply the mask to the center body of the image (more efficient)
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        for(int j = 1; j < nx-1; j++)
        {
            double sum = 0;
            for(int l = 0; l < 3; l++)
            {
                for(int m = 0; m < 3; m++)
                {
                    int p = (i + l -1) * nx + j + m -1;
                    sum += input[p] * mask[l * 3 + m];
                }
            }
            int k = i * nx + j;
            output[k] = sum;
        }
    }

    //apply the mask to the first and last rows
    #pragma omp for schedule(dynamic) nowait
    for(int j = 1; j < nx-1; j++)
    {
        double sum = 0;
        sum += input[j-1] * (mask[0] + mask[3]);
        sum += input[ j ] * (mask[1] + mask[4]);
        sum += input[j+1] * (mask[2] + mask[5]);

        sum += input[nx + j-1] * mask[6];
        sum += input[nx +  j ] * mask[7];
        sum += input[nx + j+1] * mask[8];

        output[j] = sum;

        sum = 0;
        sum += input[(ny-2)*nx+j-1] * mask[0];
        sum += input[(ny-2)*nx+j  ] * mask[1];
        sum += input[(ny-2)*nx+j+1] * mask[2];

        sum += input[(ny-1)*nx+j-1] * (mask[6] + mask[3]);
        sum += input[(ny-1)*nx+j  ] * (mask[7] + mask[4]);
        sum += input[(ny-1)*nx+j+1] * (mask[8] + mask[5]);

        output[(ny-1)*nx + j] = sum;
    }

    //apply the mask to the first and last columns
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        double sum = 0;
        sum += input[(i - 1)*nx]   * (mask[0] + mask[1]);
        sum += input[(i - 1)*nx+1] * mask[2];

        sum += input[i * nx]   * (mask[3] + mask[4]);
        sum += input[i * nx+1] * mask[5];

        sum += input[(i + 1)*nx]   * (mask[6] + mask[7]);
        sum += input[(i + 1)*nx+1] * mask[8];

        output[i*nx] = sum;

        sum = 0;
        sum += input[i * nx-2] * mask[0];
        sum += input[i * nx-1] * (mask[1] + mask[2]);

        sum += input[(i + 1)*nx-2] * mask[3];
        sum += input[(i + 1)*nx-1] * (mask[4] + mask[5]);

        sum += input[(i + 2)*nx-2] * mask[6];
        sum += input[(i + 2)*nx-1] * (mask[7] + mask[8]);

        output[i*nx + nx -1] = sum;
    }

    //apply the mask to the four corners
    #pragma omp sections
    {
    #pragma omp section
    output[0] = input[0]    * (mask[0] + mask[1] + mask[3] + mask[4]) +
                input[1]    * (mask[2] + mask[5]) +
                input[nx]   * (mask[6] + mask[7]) +
                input[nx+1] * mask[8];

    #pragma omp section
    output[nx-1] =
                input[nx-2]   * (mask[0] + mask[3]) +
                input[nx-1]   * (mask[1] + mask[2] + mask[4] + mask[5]) +
                input[2*nx-2] * mask[6] +
                input[2*nx-1] * (mask[7] + mask[8]);

    #pragma omp section
    output[(ny-1)*nx] =
                input[(ny-2)*nx]   * (mask[0] + mask[1]) +
                input[(ny-2)*nx+1] *  mask[2] +
                input[(ny-1)*nx]   * (mask[3] + mask[4] + mask[6] + mask[7]) +
                input[(ny-1)*nx+1] * (mask[5] + mask[8]);

    #pragma omp section
    output[ny*nx-1] =
                input[(ny-1)*nx-2] * mask[0] +
                input[(ny-1)*nx-1] * (mask[1] + mask[2]) +
                input[ny*nx-2] * (mask[3] + mask[6]) +
                input[ny*nx-1] * (mask[4] + mask[5] + mask[7] + mask[8]);

    }
}
}

template<class T>
inline void Dxx(const T &I, T &Ixx, const int nx, const int ny)
{
    //computing the temporal derivative
    float M[]  = {0., 0., 0.,
                  1.,-2., 1.,
                  0., 0., 0.};

    mask3x3(I, Ixx, nx, ny, M);
}

template<class T>
inline void Dyy(const T &I, T &Iyy, const int nx, const int ny)
{
    //computing the temporal derivative
    float M[]  = {0., 1., 0.,
                  0.,-2., 0.,
                  0., 1., 0.};

    mask3x3(I, Iyy, nx, ny, M);
}


template<class T>
inline void Dxy(const T &I, T &Ixy, const int nx, const int ny)
{
    //computing the temporal derivative
    float M[]  = {1./4., 0.,-1./4.,
                  0.,    0., 0.,
                 -1./4., 0., 1./4.};

    mask3x3(I, Ixy, nx, ny, M);
}

template<class T>
inline void laplacian(const T &input, T &output, const int nx, const int ny)
{
    const float a =  1.0;
    const float b = -8.0;

    float M[] = { a, a, a,
                  a, b, a,
                  a, a, a};

    mask3x3(input, output, nx, ny, M);
}


template<class T>
inline void smooth(const T &input, T &output, const int nx, const int ny)
{
    const float a = 0.07842776544;
    const float b = 0.1231940459;
    const float c = 0.1935127547;

    float M[] = { a, b, a,
                  b, c, b,
                  a, b, a};

    mask3x3(input, output, nx, ny, M);
}


#endif
