
// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Javier Sánchez Pérez <jsanchez@dis.ulpgc.es>
// All rights reserved.

#ifndef ROBUST_OPTIC_FLOW_H
#define ROBUST_OPTIC_FLOW_H

#include <omp.h>
#include <vector>

#include "mask.h"
#include "zoom.h"
#include "bilinear_interpolation.h"


#define EPSILON 0.001


void phip_smooth(
  const float *ux, 
  const float *uy, 
  const float *vx, 
  const float *vy,
  float *phi, 
  int size
)
{
#pragma omp parallel
{
    #pragma omp for
    for(int i = 0; i < size; i++)
    {
        const float du = ux[i] * ux[i] + uy[i] * uy[i];
        const float dv = vx[i] * vx[i] + vy[i] * vy[i];
        const float d2 = du + dv;

        phi[i] = 1. / sqrt(d2 + EPSILON * EPSILON);
    }
}
}


void phip_gradient(
  const float *I1x, 
  const float *I1y, 
  const float *I2x, 
  const float *I2y,
  float *phi, 
  int nx, 
  int ny
)
{
#pragma omp parallel
{
    #pragma omp for
    for(int i = 0; i < nx * ny; i++)
    {
        const float d1 = (I1x[i] - I2x[i]) * (I1x[i] - I2x[i]);
        const float d2 = (I1y[i] - I2y[i]) * (I1y[i] - I2y[i]);
        const float d3 = d1 + d2;

        phi[i] = 1. / sqrt(d3 + EPSILON * EPSILON);
    }
}
}


void divergence(float *d, float *phi, float *output, int nx, int ny)
{
 //this algorithm follows the philosophy of the mask procedure (more efficient)
 #pragma omp parallel
 {
    //apply the mask to the center body of the image
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        for(int j = 1; j < nx-1; j++)
        {
            const int k = i * nx + j;

            const float s1 = 0.5 * (phi[k +nx] + phi[k]) * d[k +nx];
            const float s2 = 0.5 * (phi[k -nx] + phi[k]) * d[k -nx];
            const float s3 = 0.5 * (phi[k + 1] + phi[k]) * d[k + 1];
            const float s4 = 0.5 * (phi[k - 1] + phi[k]) * d[k - 1];

            output[k] = s1 + s2 + s3 + s4;
        }
    }

    //apply the mask to the first and last rows
    #pragma omp for schedule(dynamic) nowait
    for(int j = 1; j < nx-1; j++)
    {
        const float s1 = 0.5 * (phi[j +nx] + phi[j]) * d[j +nx];
        const float s2 = 0.5 * (phi[j    ] + phi[j]) * d[j    ];
        const float s3 = 0.5 * (phi[j + 1] + phi[j]) * d[j + 1];
        const float s4 = 0.5 * (phi[j - 1] + phi[j]) * d[j - 1];

        output[j] = s1 + s2 + s3 + s4;

        const int   k  = (ny-1)*nx + j;
        const float s5 = 0.5 * (phi[k    ] + phi[k]) * d[k    ];
        const float s6 = 0.5 * (phi[k -nx] + phi[k]) * d[k -nx];
        const float s7 = 0.5 * (phi[k + 1] + phi[k]) * d[k + 1];
        const float s8 = 0.5 * (phi[k - 1] + phi[k]) * d[k - 1];

        output[k] = s5 + s6 + s7 + s8;
    }

    //apply the mask to the first and last columns
    #pragma omp for schedule(dynamic) nowait
    for(int i = 1; i < ny-1; i++)
    {
        const int   k  = i*nx;
        const float s1 = 0.5 * (phi[k +nx] + phi[k]) * d[k +nx];
        const float s2 = 0.5 * (phi[k -nx] + phi[k]) * d[k -nx];
        const float s3 = 0.5 * (phi[k + 1] + phi[k]) * d[k + 1];
        const float s4 = 0.5 * (phi[k    ] + phi[k]) * d[k    ];

        output[k] = s1 + s2 + s3 + s4;

        const int   j  = (i+1) * nx - 1;
        const float s5 = 0.5 * (phi[j +nx] + phi[j]) * d[j +nx];
        const float s6 = 0.5 * (phi[j -nx] + phi[j]) * d[j -nx];
        const float s7 = 0.5 * (phi[j    ] + phi[j]) * d[j    ];
        const float s8 = 0.5 * (phi[j - 1] + phi[j]) * d[j - 1];

        output[j] = s5 + s6 + s7 + s8;
    }
 }

    //apply the mask to the four corners
    #pragma omp sections
    {
    #pragma omp section
    output[0] = 2.0 * (phi[0]          ) * d[0] +
                0.5 * (phi[1 ] + phi[0]) * d[1] +
                0.5 * (phi[nx] + phi[0]) * d[nx];

    #pragma omp section
    output[nx-1] =
                2.0 * (phi[nx-1]              ) * d[nx-1] +
                0.5 * (phi[nx-2]   + phi[nx-1]) * d[nx-2] +
                0.5 * (phi[2*nx-1] + phi[nx-1]) * d[2*nx-1];

    #pragma omp section
    output[(ny-1)*nx] =
                2.0 * (phi[(ny-1)*nx]              )      * d[(ny-1)*nx]   +
                0.5 * (phi[(ny-1)*nx+1] + phi[(ny-1)*nx]) * d[(ny-1)*nx+1] +
                0.5 * (phi[(ny-2)*nx]   + phi[(ny-1)*nx]) * d[(ny-2)*nx];

    #pragma omp section
    output[ny*nx-1] =
                2.0 * (phi[ny*nx-1]                   ) * d[ny*nx-1] +
                0.5 * (phi[ny*nx-2]     + phi[ny*nx-1]) * d[ny*nx-2] +
                0.5 * (phi[(ny-1)*nx-1] + phi[ny*nx-1]) * d[ny*nx-1];
    }
}


//optical flow with the Charbonier robust functional in the data and smoothness terms
//Data term: brightness constancy term + gradient constancy term
void robust_optic_flow
(
    float *I1, float *I2, float *u, float *v, int nx, int ny,
    float dt = 10, float alpha = 1, float gamma = 0.5, float TOL = 0.01,
    int inner_iter = 10, int outer_iter = 80
)
{
    const int size = nx * ny;

    float *uk   = new float[size];
    float *vk   = new float[size];
    float *u_k  = new float[size];
    float *v_k  = new float[size];
    float *ux   = new float[size];
    float *uy   = new float[size];
    float *vx   = new float[size];
    float *vy   = new float[size];
    float *I1x  = new float[size];
    float *I1y  = new float[size];
    float *I2x  = new float[size];
    float *I2y  = new float[size];
    float *I2w  = new float[size];
    float *I2wx = new float[size];
    float *I2wy = new float[size];
    float *I2xx = new float[size];
    float *I2yy = new float[size];
    float *I2xy = new float[size];
    float *I2wxx= new float[size];
    float *I2wyy= new float[size];
    float *I2wxy= new float[size];
    float *phi  = new float[size];
    float *phid = new float[size];
    float *phiu = new float[size];
    float *phiv = new float[size];
    float *phig = new float[size];
    float *mask = new float[size];

    //compute the gradient of the images
    gradient(I1, I1x, I1y, nx, ny);
    gradient(I2, I2x, I2y, nx, ny);

    //compute second derivatives
    Dxx(I2, I2xx, nx, ny);
    Dyy(I2, I2yy, nx, ny);
    Dxy(I2, I2xy, nx, ny);

    for(int i = 0; i < outer_iter; i++)
    {
        std::copy(u, &u[size], u_k);
        std::copy(v, &v[size], v_k);

        bilinear_interpolation(I2,  u, v, I2w,  nx, ny, mask);
        bilinear_interpolation(I2x, u, v, I2wx, nx, ny, mask);
        bilinear_interpolation(I2y, u, v, I2wy, nx, ny, mask);

        bilinear_interpolation(I2xx, u, v, I2wxx, nx, ny, mask);
        bilinear_interpolation(I2xy, u, v, I2wxy, nx, ny, mask);
        bilinear_interpolation(I2yy, u, v, I2wyy, nx, ny, mask);

        phip_gradient(I1x, I1y, I2wx, I2wy, phig, nx, ny);

        for(int n = 0; n < inner_iter; n++)
        {
            gradient(u, ux, uy, nx, ny);
            gradient(v, vx, vy, nx, ny);

            phip_smooth(ux, uy, vx, vy, phi, size);

            float M[] = {0. , 0.5, 0.,
                         0.5, 2. , 0.5,
                         0. , 0.5, 0.};

            mask3x3(phi, phid, nx, ny, M);

            divergence(u, phi, phiu, nx, ny);
            divergence(v, phi, phiv, nx, ny);

            std::copy(u, &u[size], uk);
            std::copy(v, &v[size], vk);

            int ng = 0;
            float error = 1000;
            while(error > TOL && ng < inner_iter)
            {
                ng++;
                error = 0;

                #pragma omp parallel reduction(+ : error)
                {
                    #pragma omp for
                    for(int i = 0; i < size; i++)
                    {
                        //brightness constancy term
                        const float I2wl = I2wx[i] * u_k[i] + I2wy[i] * v_k[i];
                        const float dif  = I1[i] - I2w[i] + I2wl;
                        const float dif2 = I1[i] - I2w[i] - I2wx[i] * (uk[i] - u_k[i]) -
                                           I2wy[i] * (vk[i] - v_k[i]);
                        const float d2   = dif2 * dif2;
                        const float phi1 = mask[i] * 1.0 / sqrt(d2 + EPSILON * EPSILON);
                        const float BNu  = phi1  * (dif - I2wy[i] * v[i]) * I2wx[i];
                        const float BNv  = phi1  * (dif - I2wx[i] * u[i]) * I2wy[i];
                        const float BDu  = phi1  * I2wx[i] * I2wx[i];
                        const float BDv  = phi1  * I2wy[i] * I2wy[i];

                        //gradient constancy term
                        const float dx  = (I1x[i] - I2wx[i]);
                        const float dy  = (I1y[i] - I2wy[i]);
                        const float du  = (uk[i] - u_k[i]);
                        const float dv  = (vk[i] - v_k[i]);
                        const float c   = mask[i] * gamma * phig[i];
                        const float GNu = c * ((dx + I2wxx[i] * u_k[i] - I2wxy[i] * dv) * I2wxx[i] +
                                               (dy + I2wxy[i] * u_k[i] - I2wyy[i] * dv) * I2wxy[i]
                                              );
                        const float GNv = c * ((dx - I2wxx[i] * du + I2wxy[i] * v_k[i]) * I2wxy[i] +
                                               (dy - I2wxy[i] * du + I2wyy[i] * v_k[i]) * I2wyy[i]
                                              );
                        const float GDu = c * (I2wxx[i] * I2wxx[i] + I2wxy[i] * I2wxy[i]);
                        const float GDv = c * (I2wyy[i] * I2wyy[i] + I2wxy[i] * I2wxy[i]);

                        //smoothness term
                        const float SNu = alpha * phiu[i];
                        const float SNv = alpha * phiv[i];
                        const float SD  = alpha * phid[i];

                        const float ukk  = u[i];
                        const float vkk  = v[i];

                        u[i] = (uk[i] + dt * (BNu + GNu + SNu))
                                / (1. + dt * (BDu + GDu + SD ));

                        v[i] = (vk[i] + dt * (BNv + GNv + SNv))
                                / (1. + dt * (BDv + GDv + SD )); 

                        error += (u[i] - ukk) * (u[i] - ukk) +
		                         (v[i] - vkk) * (v[i] - vkk);

                    }

                    error /= size;
                }
            }
        }
    }

    delete []uk;
    delete []vk;
    delete []u_k;
    delete []v_k;
    delete []ux;
    delete []uy;
    delete []vx;
    delete []vy;
    delete []I1x;
    delete []I1y;
    delete []I2x;
    delete []I2y;
    delete []I2w;
    delete []I2wx;
    delete []I2wy;
    delete []I2xx;
    delete []I2yy;
    delete []I2xy;
    delete []I2wxx;
    delete []I2wyy;
    delete []I2wxy;
    delete []phi;
    delete []phid;
    delete []phiu;
    delete []phiv;
    delete []phig;
    delete []mask;
}


void optic_flow_zoom_robust(float *I1, float *I2, float *u, float *v, int nxx, int nyy,
                            float dt = 5, float alpha = 10, float gamma = 0.5, int nscales = 5,
                            float zoom_factor = 0.5, float TOL = 0.01,
                            int inner_iter = 10, int outer_iter = 30, bool verbose = false)
{
    int size = nxx * nyy;

    std::vector<float *> I1s(nscales);
    std::vector<float *> I2s(nscales);
    std::vector<float *> us (nscales);
    std::vector<float *> vs (nscales);

    std::vector<int> nx(nscales);
    std::vector<int> ny(nscales);

    I1s[0] = I1;
    I2s[0] = I2;

    //presmoothing the finest scale images
    gaussian(I1s[0], nxx, nyy, 0.8);
    gaussian(I2s[0], nxx, nyy, 0.8);

    us [0] = u;
    vs [0] = v;
    nx [0] = nxx;
    ny [0] = nyy;

    std::fill_n(us[0], size, 0.0);
    std::fill_n(vs[0], size, 0.0);

    //create the scales
    for(int s = 1; s < nscales; s++)
    {
        zoom_size(nx[s-1], ny[s-1], nx[s], ny[s], zoom_factor);
        const int sizes = nx[s] * ny[s];

        I1s[s] = new float[sizes];
        I2s[s] = new float[sizes];
        us[s]  = new float[sizes];
        vs[s]  = new float[sizes];

        //compute the zoom from the previous scale
        zoom(I1s[s-1], I1s[s], nx[s-1], ny[s-1], zoom_factor);
        zoom(I2s[s-1], I2s[s], nx[s-1], ny[s-1], zoom_factor);

	std::fill_n(us[s], sizes, 0.0);
        std::fill_n(vs[s], sizes, 0.0);
    }

    for(int s = nscales-1; s >= 0; s--)
    {
        if(verbose) std::cout << "Scale: " << s << std::endl;

        robust_optic_flow(
                I1s[s], I2s[s], us[s], vs[s],
                nx[s], ny[s], dt, alpha, gamma,
                TOL, inner_iter, outer_iter
        );

        if(s)
        {
            zoom_out(us[s], us[s-1], nx[s], ny[s], nx[s-1], ny[s-1]);
            zoom_out(vs[s], vs[s-1], nx[s], ny[s], nx[s-1], ny[s-1]);

            int size = nx[s-1] * ny[s-1];

            std::transform(us[s-1], &us[s-1][size], us[s-1],
                           std::bind2nd(std::multiplies<float>(),
                           1.0 / zoom_factor));

            std::transform(vs[s-1], &vs[s-1][size], vs[s-1],
                           std::bind2nd(std::multiplies<float>(),
                           1.0 / zoom_factor));
        }
    }

    for(int i = 1; i < nscales; i++)
    {
        delete []I1s[i];
        delete []I2s[i];
        delete []us [i];
        delete []vs [i];
    }
}


#endif
