
// This program is free software: you can use, modify and/or redistribute it
// under the terms of the simplified BSD License. You should have received a
// copy of this license along this program. If not, see
// <http://www.opensource.org/licenses/bsd-license.html>.
//
// Copyright (C) 2012, Javier Sánchez Pérez <jsanchez@dis.ulpgc.es>
// All rights reserved.

#ifndef ZOOM_H
#define ZOOM_H

#include <omp.h>
#include <math.h>

#include "mask.h"
#include "bicubic_interpolation.h"
#include "gaussian.h"



/**
  *
  * Calculate the size of the zoomed image
  *
**/
void zoom_size(
    int nx,             ///width of the orignal image
    int ny,             ///height of the orignal image
    int &nxx,           ///width of the zoomed image
    int &nyy,           ///height of the zoomed image
    float factor = 0.5  ///zoom factor between 0 and 1
)
{
    nxx = (int)((float) nx * factor + 0.5);
    nyy = (int)((float) ny * factor + 0.5);
}


/**
  *
  * Function to zoom in the image
  *
**/
void zoom(
    const float *I,          ///input image
    float *Iout,             ///output image
    const int nx,            ///image width
    const int ny,            ///image height
    const float factor = 0.5 ///zoom factor between 0 and 1
)
{
    int nxx, nyy;

    float *Is = new float[nx * ny];

    for(int i = 0; i < nx * ny; i++)
        Is[i] = I[i];

    zoom_size(nx, ny, nxx, nyy, factor);

    ///compute the Gaussian sigma for smoothing
    const float sigma = 0.6 * sqrt(1.0/(factor*factor) - 1.0);

    ///pre-smooth the image
    gaussian(Is, nx, ny, sigma);

    #pragma omp parallel
    {
    #pragma omp for schedule(dynamic) nowait
    for (int i = 0; i < nyy; i++)

       for (int j = 0; j < nxx; j++)
        {
            const float x  = ((float)j/factor);
            const float y  = ((float)i/factor);

            float m;

            Iout[i * nxx + j] = bicubic_interpolation(Is, x, y, nx, ny, m);
        }
    }

    delete []Is;
}



/**
  *
  * Function to zoom out the image
  *
**/
void zoom_out(
    const float *I, ///input image
    float *Iout,    ///output image
    int nx,         ///width of the original image
    int ny,         ///height of the original image
    int nxx,        ///width of the zoomed image
    int nyy         ///height of the zoomed image
)
{
    const float factorx = ((float)nxx / nx);
    const float factory = ((float)nyy / ny);

 #pragma omp parallel
 {
    #pragma omp for schedule(dynamic) nowait
    for (int i = 0; i < nyy; i++)

        for (int j = 0; j < nxx; j++)
        {
            int l =  (int) ((float)i/factory);
            int k =  (int) ((float)j/factorx);

            Iout[i * nxx + j] = I[k + nx * l];
        }
 }
}



#endif
